/* Topology Builder        Bruce D. Ray
   IUPUI Physics Dept. NMR Center
   402 N. Blackford St.
   Indianapolis, IN   46202
   USA
        Functions to read antechamber atom types and atom correlation
        data files and to free the storage allocated in reading them.

   NOTICE: This is a derivative work based on study of the routines
   found in the antechamber 1.27 program atomtype, version 1.0,
   dated October, 2001 by Junmei Wang, Department of Pharmaceutical
   Chemistry, School of Pharmacy, University of California,
   San Francisco, CA  94143

   Portions of this work simplify storage allocation, and clarify
   decision trees compared to the work cited above.
*/

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "block_memory.h"
#include "atom_types.h"
#include "readmol2.h"

#define MAXCHAR 256
#define MAXNAME 128

/* read an antechamber style atom type definition file
   parameters are:
        at_type_cnt                   returned count of atom types
        wild_cnt                      returned count of wild atoms
        max_type                      maximum number of atom types allowed
        max_wild                      maximum number of wild atoms allowed
        just_count                    count records and return the counts
                                      but do not allocate storage or store data
        filename                      name of the atom type definition file
                                      to read
*/
void read_at_types(int *at_type_cnt, int *wild_cnt, int max_type, int max_wild,
                   int just_count, char *filename)
{
     int i, j, k, num_wild, num_type, wild_allow;
     float dummy;
     FILE *def;
     char mess[MAXCHAR], name[11][MAXNAME], line[2*MAXCHAR];
     char atype[5];

     if(just_count) {
          wild_name = (char **)mat_alloc(max_wild, 5, FARGS);
          wild_count = (int *)allocator(max_wild, sizeof(int), FARGS);
          at_type_name = (char **)mat_alloc(max_type, 12, FARGS);
          at_type_residue = (int *)allocator(max_type, sizeof(int), FARGS);
          residue_name = (char **)allocator(max_type, sizeof(char *), FARGS);
          at_type_no = (int *)allocator(max_type, sizeof(int), FARGS);
          at_type_attached = (int *)allocator(max_type, sizeof(int), FARGS);
          at_type_attached_H = (int *)allocator(max_type, sizeof(int), FARGS);
          at_type_ewd = (int *)allocator(max_type, sizeof(int), FARGS);
          at_type_prop = (char **)mat_alloc(max_type, 36, FARGS);
          at_type_env = (char **)mat_alloc(max_type, 55, FARGS);
          at_env_bonds = (char **)mat_alloc(max_type, 36, FARGS);
          wild_elements = (char ***)allocator(max_wild, sizeof(char*), FARGS);
          wild_atno = (int **)allocator(max_wild, sizeof(int*), FARGS);
          for(i = 0; i < max_type; i++) {
               at_type_residue[i] = 0;
               at_type_no[i] = 0;
               at_type_attached[i] = 999;
               at_type_attached_H[i] = 999;
               at_type_ewd[i] = 999;
               at_type_prop[i][0] = NULL;
               at_type_env[i][0] = NULL;
               at_env_bonds[i][0] = NULL;
          }
          for(i = 0; i < max_wild; i++)
               wild_count[i] = 0;
     }

     num_wild = 0;
     num_type = 0;

     if((def = fopen(filename, "r")) == NULL) {
          sprintf(mess, "Cannot open file %s\n", filename);
          my_fatal(FARGS, mess);
          exit(1);
     }

     wild_allow = 1;

     while(fgets(line, 2*MAXCHAR, def) !=NULL) {
          if((!strncmp(line, "WILDATOM", 8)) && wild_allow) {
               if(num_wild >= max_wild) {
                    sprintf(mess, "More WILDATOM records (%d) than allowed (%d)\n",
                            (num_wild+1), max_wild);
                    my_fatal(FARGS, mess);
                    exit(1);
               }
               if((i = sscanf(&line[9], "%4s %4s %4s %4s %4s %4s %4s %4s %4s %4s %4s", 
                              &name[0][0], &name[1][0], &name[2][0], &name[3][0],
                              &name[4][0], &name[5][0], &name[6][0], &name[7][0],
                              &name[8][0], &name[9][0], &name[10][0])) < 2) {

                    my_fatal(FARGS, "Faulty WILDATOM definition\n");
                    exit(1);
               }

/* only allocate the memory I need and store a count of elements for each WILDATOM */

               if(just_count) {
                    strcpy(&wild_name[num_wild][0], &name[0][0]);
                    wild_elements[num_wild] = (char **)mat_alloc((i-1), 5, FARGS);
                    wild_atno[num_wild] = (int *)allocator((i-1), sizeof(int), FARGS);
                    wild_count[num_wild] = i-1;
                    for(j = 0; j < (i-1); j++) {
                         strcpy(&wild_elements[num_wild][j][0], &name[(j+1)][0]);
                         k = strcspn(&name[(j+1)][0], "0123456789");
                         strncpy(&atype[0], &name[(j+1)][0], k);
                         atype[k] = NULL;
                         set_atomic_no(k, &atype[0], &dummy, &wild_atno[num_wild][j],
                                       mess);
                    }
               }
               num_wild++;
               continue;
          }                    /* end if((!strncmp(line, "WILDATOM", 8)) && wild_allow) */

          if(!strncmp(line, "ATD", 3)) {           /* atom type definitions */
               wild_allow = 0;                     /* no more WILDATOM entries */
               if(num_type >= max_type) {
                    sprintf(mess, "More atom type definitions (%d) than allowed (%d)\n",
                            (num_type+1), max_type);
                    my_fatal(FARGS, mess);
                    exit(1);
               }

               i = sscanf(&line[4], "%s%s%s%s%s%s%s%s%s", 
                          &name[2][0], &name[3][0], &name[4][0], &name[5][0],
                          &name[6][0], &name[7][0], &name[8][0], &name[9][0],
                          &name[10][0]);
               if((!i) || (!strcmp(&name[2][0], "&"))) {
                    my_fatal(FARGS, "Faulty ATD line\n");
                    exit(1);
               }
               num_type++;
               if(just_count) {
                    strncpy(&at_type_name[(num_type - 1)][0], &name[2][0], 10);
                    at_type_name[(num_type - 1)][10] = NULL;
                    if(!strcmp(&name[3][0], "&")) continue;
                    at_type_residue[(num_type - 1)] = -1;
                    if(strcmp(&name[3][0], "*")) {
                         if(!strcmp(&name[3][0], "AA"))
                              at_type_residue[(num_type - 1)] = 1;
                         if(!strcmp(&name[3][0], "NA"))
                              at_type_residue[(num_type - 1)] = 2;
                         if(!strcmp(&name[3][0], "BIO"))
                              at_type_residue[(num_type - 1)] = 3;
                         if(strcmp(&name[3][0], "AA") && strcmp(&name[3][0], "NA") &&
                            strcmp(&name[3][0], "BIO")) {
                                at_type_residue[(num_type - 1)] = 4;
                                residue_name[(num_type - 1)] =
                                     (char *)allocator((strlen(&name[3][0]) + 2), sizeof(char), FARGS);
                                strcpy(&residue_name[(num_type - 1)][0], &name[3][0]);
                         }
                    }                               /* end if(strcmp(&name[3][0], "*")) */
                    if(!strcmp(&name[4][0], "&")) continue;
                    if(strcmp(&name[4][0], "*")) {
                         if(strspn(&name[4][0], "0123456789") != strlen(&name[4][0])) {
                              my_fatal(FARGS, "Faulty atomic number\n");
                              exit(1);
                         }
                         at_type_no[(num_type - 1)] = atoi(&name[4][0]);
                    }
                    else at_type_no[(num_type - 1)] = -1;
                    if(!strcmp(&name[5][0], "&")) continue;
                    if(strcmp(&name[5][0], "*")) {
                         if(strspn(&name[5][0], "0123456789") != strlen(&name[5][0])) {
                              my_fatal(FARGS, "Faulty number of attached atoms\n");
                              exit(1);
                         }
                         at_type_attached[(num_type - 1)] = atoi(&name[5][0]);
                    }
                    else at_type_attached[(num_type - 1)] = -1;
                    if(!strcmp(&name[6][0], "&")) continue;
                    if(strcmp(&name[6][0], "*")) {
                         if(strspn(&name[6][0], "0123456789") != strlen(&name[6][0])) {
                              my_fatal(FARGS, "Faulty H attachments number\n");
                              exit(1);
                         }
                         at_type_attached_H[(num_type - 1)] = atoi(&name[6][0]);
                    }
                    else at_type_attached_H[(num_type - 1)] = -1;
                    if(!strcmp(&name[7][0], "&")) continue;
                    if(strcmp(&name[7][0], "*")) {
                         if(strspn(&name[7][0], "0123456789") != strlen(&name[7][0])) {
                              my_fatal(FARGS, "Faulty electron withdrawal number\n");
                              exit(1);
                         }
                         at_type_ewd[(num_type - 1)] = atoi(&name[7][0]);
                    }
                    else at_type_ewd[(num_type - 1)] = -1;
                    if(!strcmp(&name[8][0], "&")) continue;
                    if(strcmp(&name[8][0], "*")) {
                         strncpy(&at_type_prop[(num_type - 1)][0], &name[8][0], 34);
                         at_type_prop[(num_type - 1)][34] = NULL;
                    }
                    else strcpy(&at_type_prop[(num_type - 1)][0], "*");
                    if(!strcmp(&name[9][0], "&")) continue;
                    if(strcmp(&name[9][0], "*")) {
                         strncpy(&at_type_env[(num_type - 1)][0], &name[9][0], 52);
                         at_type_env[(num_type - 1)][52] = NULL;
                    }
                    else strcpy(&at_type_env[(num_type - 1)][0], "*");
                    if(!strcmp(&name[10][0], "&")) continue;
                    strcpy(&at_env_bonds[(num_type - 1)][0], &name[10][0]);
               }                   /* end if(just_count) */
          }                        /* end if(!strncmp(line, "ATD", 3)) */
     }                             /* end while(fgets(line, 2*MAXCHAR, def) 1=NULL) */

     *wild_cnt = num_wild;
     num_wilds = num_wild;
     *at_type_cnt = num_type;
     fclose(def);

     return;
}

/* read an antechamber style atom types correlation file
   parameters are:
        corr_cnt                  returned count of number of atom types
                                  correlation records
        max_corr                  maximum number of atom types correlations
                                  allowed
        just_count                count records and return the counts
                                  but do not allocate storage or store data
        filename                  name of the atom types correlations file
                                  to read
*/
void read_at_corr(int *corr_cnt, int max_corr, int just_count,
                  char *filename)
{
     int i, j, k, count;
     char mess[MAXCHAR], line[MAXCHAR], name[5][6];
     FILE *corr;

     if(just_count) {
          corr_name = (char **)mat_alloc(max_corr, 4, FARGS);
          index_improper = (int *)allocator(max_corr, sizeof(int), FARGS);
          corr_to = (char ***)tabl_alloc(max_corr, 3, 4, FARGS);
          corr_num = (int *)allocator(max_corr, sizeof(int), FARGS);
          for(i = 0; i < max_corr; i++)
               corr_num[i] = 0;
     }

     if((corr = fopen(filename, "r")) == NULL) {
          sprintf(mess, "Cannot open file %s\n", filename);
          my_fatal(FARGS, mess);
          exit(1);
     }

     count = 0;
     while(fgets(line, MAXCHAR, corr) !=NULL) {
          if(!strncmp(line, "CORR", 4)) {
               if((i = sscanf(&line[4], "%s%s%s%s%s", name[0], name[1],
                              name[2], name[3], name[4])) < 2) {

                    my_fatal(FARGS, "Faulty atom correlation definition\n");
                    exit(1);
               }
               if(i == 3)
                    if(!strcmp(name[0], name[2])) {
                         name[2][0] = NULL;
                         i = 2;
                    }
               for(k = 0; k < i; k++) {                          /* append spaces to names as needed */
                    if(k == 1) continue;
                    if(strlen(name[k]) == 1) {
                         name[k][1] = ' ';
                         name[k][2] = NULL;
                    }
               }

               if(just_count) {
                    strncpy(&corr_name[count][0], &name[0][0], 3);
                    corr_name[count][3] = NULL;
                    corr_num[count] = i - 2;
                    if(strspn(&name[1][0], "0123456789") != strlen(&name[1][0])) {
                         my_fatal(FARGS, "Faulty atom impropers index\n");
                         exit(1);
                    }
                    index_improper[count] = atoi(name[1]);
                    for(j = 0; j < (i-2); j++) {
                         strncpy(&corr_to[count][j][0], &name[(j+2)][0], 3);
                         corr_to[count][j][3] = NULL;
                    }
               }                   /* end if(just_count) */
          count++;
          }                        /* end if(!strncmp(line, "CORR", 4)) */
     }                             /* end while(fgets(line, MAXCHAR, corr) !=NULL) */

     *corr_cnt = count;
     fclose(corr);
     return;
}

/* free storage allocated for atom types data
   parameters are:
        at_type_cnt               number of atom types
        wild_cnt                  number of wild atom records
*/
void release_types(int at_type_cnt, int wild_cnt)
{
     int i, j;

/* free wilds arrays */
     for(i = 0; i < wild_cnt; i++){
          for(j = 0; j < wild_count[i]; j++)
               free(wild_elements[i][j]);
          free(wild_name[i]);
          free(wild_elements[i]);
          free(wild_atno[i]);
     }
     free(wild_name);
     free(wild_elements);
     free(wild_atno);

     free(wild_count);

/* free atom type arrays */
     free(at_type_residue);
     free(at_type_no);
     free(at_type_attached);
     free(at_type_attached_H);
     free(at_type_ewd);

     for(i = 0; i < at_type_cnt; i++) {
          free(at_type_name[i]);
          free(at_type_prop[i]);
          free(at_type_env[i]);
          free(at_env_bonds[i]);
     }
     free(at_type_name);
     free(at_type_prop);
     free(at_type_env);
     free(at_env_bonds);

     return;
}

/* free storage allocated for atom types correlations data
   parameter is:
        corr_count                number of atom types correlations stored
*/
void release_corr(int corr_count)
{
     int i, j;

     for(i = 0; i < corr_count; i++) {
          free(corr_name[i]);
          free(corr_to[i][0]);
          free(corr_to[i][1]);
          free(corr_to[i][2]);
          free(corr_to[i]);
     }
     free(corr_to);
     free(corr_name);
     free(index_improper);
     free(corr_num);

     return;
}
