/* Topology Builder        Bruce D. Ray
   IUPUI Physics Dept. NMR Center
   402 N. Blackford St.
   Indianapolis, IN   46202
   USA
        force field comparison routines compare_FF.c

   NOTICE: This is a derivative work based on study of the routines
   found in the antechamber 1.27 program parmchk, version 1.0,
   dated October, 2001 by Junmei Wang, Department of Pharmaceutical
   Chemistry, School of Pharmacy, University of California,
   San Francisco, CA  94143

   Portions of this work simplify storage allocation, and clarify
   decision trees compared to the work cited above.
*/

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "compare_FF.h"

/* comparison of bond to force field
   parameters are:
        atom types of each atom in bond:                  bond_atom1
                                                          bond_atom2
        number of bond parameter sets in force field:     bind_cntFF
        bond force returned:                              force
        bond length returned:                             length
*/
int compare_bonds(char *bond_atom1, char *bond_atom2, int bind_cntFF,
                  float *force, float *length)
{
     int j;

     for(j = 0; j < bind_cntFF; j++) {
          if(!((!strcmp(bond_atom1, bond_name1[j])   &&
                !strcmp(bond_atom2, bond_name2[j]))  ||
               (!strcmp(bond_atom1, bond_name2[j])   &&
                !strcmp(bond_atom2, bond_name1[j])))) continue;

          *force  = bond_force[j];
          *length = bond_length[j];
          return(1);
     }

     return(0);
}

/* comparison of angle to force field
   parameters are:
        atom types of each atom in angle:                 angle_atom1
                                                          angle_atom2
                                                          angle_atom3
        number of angle parameter sets in force field:    angle_cntFF
        angle force returned:                             force
        angle value returned:                             angle
*/
int compare_angles(char *angle_atom1, char *angle_atom2, char *angle_atom3,
                   int angle_cntFF, float *force, float *angle, int *method)
{
     int j, k;

     k = *method;
     for(j = 0; j < angle_cntFF; j++) {
          if(strcmp(angle_atom2, angle_name2[j])) continue;          /* central atom must always match */

          if(!((!strcmp(angle_atom1, angle_name1[j])  &&
                !strcmp(angle_atom3, angle_name3[j])) ||
               (!strcmp(angle_atom1, angle_name3[j])  &&
                !strcmp(angle_atom3, angle_name1[j])))) continue;

          *force = angle_force[j];
          *angle = angle_angle[j];
          *method = k + 1;                            /* signal found in table */
          return(1);
     }

     return(0);
}

/* empirical method to calculate angle phase and force 
   parameters are:
        atom index numbers of angle atoms:                ang1, ang2, ang3
        computed angle force returned:                    force
        computed angle value returned:                    angle
        molecule bond lengths to use already set:         bnd_length
        similar atoms index pointer:                      sim_atom_indx
        similar atoms list pointers:                      corr_name1
                                                          corr_name2
                                                          corr_name3
        number of angle parameter sets in force field:    angle_cntFF
        assigned atom types to use:                       assigned_type
*/
void calc_angle(int ang1, int ang2, int ang3, float *force, float *angle,
                float *bnd_length, int *method, int *sim_atom_indx,
                char **corr_name1, char **corr_name2, char **corr_name3,
                int angle_cntFF, char **assigned_type)
{
     int i, k, brkr;
     double length1, length2, my_angle, my_force, param[4];
     float angle1, angle2, use_force;

     brkr = 0;
     while(!brkr) {                             /* device to allow me to skip things without */
                                                /* a really nast nesting of if's */
          if((brkr = compare_angles(assigned_type[ang1], assigned_type[ang2],
                                    assigned_type[ang1], angle_cntFF,
                                    &use_force, &angle1, &k))) continue;

          if(corr_name1 == NULL) break;                /* flag lack of alternate names */

          if(sim_atom_indx[ang1] > 1)
               if((brkr = compare_angles(corr_name2[ang1], assigned_type[ang2],
                                         corr_name2[ang1], angle_cntFF,
                                         &use_force, &angle1, &k))) continue;

          if(sim_atom_indx[ang2] > 1)
               if((brkr = compare_angles(assigned_type[ang1], corr_name2[ang2],
                                         assigned_type[ang1], angle_cntFF,
                                         &use_force, &angle1, &k))) continue;

          if((sim_atom_indx[ang1] > 1) && (sim_atom_indx[ang2] > 1))
               if((brkr = compare_angles(corr_name2[ang1], corr_name2[ang2],
                                         corr_name2[ang1], angle_cntFF,
                                         &use_force, &angle1, &k))) continue;

          if(sim_atom_indx[ang1] > 2)
               if((brkr = compare_angles(corr_name3[ang1], assigned_type[ang2],
                                         corr_name3[ang1], angle_cntFF,
                                         &use_force, &angle1, &k))) continue;

          if(sim_atom_indx[ang2] > 2)
               if((brkr = compare_angles(assigned_type[ang1], corr_name3[ang2],
                                         assigned_type[ang1], angle_cntFF,
                                         &use_force, &angle1, &k))) continue;

          if((sim_atom_indx[ang1] > 2) && (sim_atom_indx[ang2] > 2))
               if((brkr = compare_angles(corr_name3[ang1], corr_name3[ang2],
                                         corr_name3[ang1], angle_cntFF,
                                         &use_force, &angle1, &k))) continue;

          if(sim_atom_indx[ang1] > 0)
               if((brkr = compare_angles(corr_name1[ang1], assigned_type[ang2],
                                         corr_name1[ang1], angle_cntFF,
                                         &use_force, &angle1, &k))) continue;

          if(sim_atom_indx[ang2] > 0)
               if((brkr = compare_angles(assigned_type[ang1], corr_name1[ang2],
                                         assigned_type[ang1], angle_cntFF,
                                         &use_force, &angle1, &k))) continue;

          if((sim_atom_indx[ang1] > 0) && (sim_atom_indx[ang2] > 1))
               if((brkr = compare_angles(corr_name1[ang1], corr_name1[ang2],
                                         corr_name1[ang1], angle_cntFF,
                                         &use_force, &angle1, &k))) continue;

          break;                                 /* failure, exit loop */
     }

     if(!brkr) return;                       /* failed to calculate angle parameters */

     brkr = 0;
     while(!brkr) {                             /* device to allow me to skip things without */
                                                /* a really nast nesting of if's */
          if((brkr = compare_angles(assigned_type[ang3], assigned_type[ang2],
                                    assigned_type[ang3], angle_cntFF,
                                    &use_force, &angle2, &k))) continue;

          if(corr_name1 == NULL) break;                /* flag lack of alternate names */

          if(sim_atom_indx[ang3] > 1)
               if((brkr = compare_angles(corr_name2[ang3], assigned_type[ang2],
                                         corr_name2[ang3], angle_cntFF,
                                         &use_force, &angle2, &k))) continue;

          if(sim_atom_indx[ang2] > 1)
               if((brkr = compare_angles(assigned_type[ang3], corr_name2[ang2],
                                         assigned_type[ang3], angle_cntFF,
                                         &use_force, &angle2, &k))) continue;

          if((sim_atom_indx[ang3] > 1) && (sim_atom_indx[ang2] > 1))
               if((brkr = compare_angles(corr_name2[ang3], corr_name2[ang2],
                                         corr_name2[ang3], angle_cntFF,
                                         &use_force, &angle2, &k))) continue;

          if(sim_atom_indx[ang3] > 2)
               if((brkr = compare_angles(corr_name3[ang3], assigned_type[ang2],
                                         corr_name3[ang3], angle_cntFF,
                                         &use_force, &angle2, &k))) continue;

          if(sim_atom_indx[ang2] > 2)
               if((brkr = compare_angles(assigned_type[ang3], corr_name3[ang2],
                                         assigned_type[ang3], angle_cntFF,
                                         &use_force, &angle2, &k))) continue;

          if((sim_atom_indx[ang3] > 2) && (sim_atom_indx[ang2] > 2))
               if((brkr = compare_angles(corr_name3[ang3], corr_name3[ang2],
                                         corr_name3[ang3], angle_cntFF,
                                         &use_force, &angle2, &k))) continue;

          if(sim_atom_indx[ang3] > 0)
               if((brkr = compare_angles(corr_name1[ang3], assigned_type[ang2],
                                         corr_name1[ang3], angle_cntFF,
                                         &use_force, &angle2, &k))) continue;

          if(sim_atom_indx[ang2] > 0)
               if((brkr = compare_angles(assigned_type[ang3], corr_name1[ang2],
                                         assigned_type[ang3], angle_cntFF,
                                         &use_force, &angle2, &k))) continue;

          if((sim_atom_indx[ang3] > 0) && (sim_atom_indx[ang2] > 1))
               if((brkr = compare_angles(corr_name1[ang3], corr_name1[ang2],
                                         corr_name1[ang3], angle_cntFF,
                                         &use_force, &angle2, &k))) continue;

          break;                                 /* failure, exit loop */
     }

     if(!brkr) return;                       /* failed to calculate angle parameters */

     my_angle = 0.5 * (angle1 + angle2);

     brkr = 0;                                  /* retrieve lengths and forces of both bonds */
     for(i = 0; i < bond_count[ang2]; i++) {
          if((bond_i[(bonds_list[ang2][i])] == ang1)  ||
             (bond_j[(bonds_list[ang2][i])] == ang1)) {
                    length1 = bnd_length[(bonds_list[ang2][i])];
                    brkr = 1;
          }

          if((bond_i[(bonds_list[ang2][i])] == ang3)  ||
             (bond_j[(bonds_list[ang2][i])] == ang3)) {
                    length2 = bnd_length[(bonds_list[ang2][i])];
                    brkr = 2;
          }
          if(brkr == 2) break;
     }

     if(brkr !=2) return;                     /* something wrong. failed to find the bonds! */
     if((length1 > 9999.0) || (length2 > 9999.0)) return;      /* something not set */

     param[1] = (length1 - length2) * (length1 - length2);
     param[1] = param[1] / ((length1 + length2) * (length1 + length2));

     switch(atom_atno[ang2]) {
          case  6: param[2] = 1.339;
                   break;

          case  7: param[2] = 1.300;
                   break;

          case  8: param[2] = 1.249;
                   break;

          case 15: param[2] = 1.448;
                   break;

          case 16: param[2] = 0.906;
                   break;

          default: param[2] = 0.0;
                   break;
     }

     param[3] = force_param(atom_atno[ang1]);
     param[4] = force_param(atom_atno[ang3]);

     my_force = 143.9 * param[3] * param[2] * param[4] * exp(-2.0 * param[1]) / (length1 + length2);
     my_force /= sqrt(my_angle * 3.1415926 / 180.0);
     *force = my_force;
     *angle = my_angle;
     *method = 9;                          /* signal empirical calculation */

     return;
}

/* based on atomic number, return a parameter to compute an angle force
   input parameter is:
        atomic number         atomic
*/
double force_param(int atomic)
{
     switch(atomic) {
          case  1: return(0.784);
                   break;

          case  6: return(1.183);
                   break;

          case  7: return(1.212);
                   break;

          case  8: return(1.219);
                   break;

          case  9: return(1.166);
                   break;

          case 15: return(1.280);
                   break;

          case 16: return(1.620);
                   break;

          case 17: return(1.272);
                   break;

          case 35: return(1.378);
                   break;

          case 53: return(1.398);
                   break;

          default: return(0.0);
                   break;
     }

     return(0.0);                           /* should never get here */
}

/* comparison of torsion to force field
   parameters are:
        atom types of each atom in torsion:               tors_atom1
                                                          tors_atom2
                                                          tors_atom3
                                                          tors_atom4
        number of torsion parameter sets in force field:  tors_cntFF
        torsion multiplier returned:                      mult
        torsion force returned:                           force
        torsion angle returned:                           phase
        torsion term returned:                            term
*/
int compare_tors(char *tors_atom1, char *tors_atom2, char *tors_atom3,
                 char *tors_atom4, int tors_cntFF, int *mult,
                 float *force, float *phase, float *term)
{
     int j;

     for(j = 0; j < tors_cntFF; j++) {
          if((!strcmp(tors_atom2, tors_name2[j]) &&
              !strcmp(tors_atom3, tors_name3[j])))                      /* two center must match */
                    if((!strcmp(tors_atom1, tors_name1[j])  &&          /* 1 goes to 2 and 4 goes to 3 */
                        !strcmp(tors_atom4, tors_name4[j])) ||
                       ((tors_name1[j][0] == 'X') &&
                        (tors_name4[j][0] == 'X'))) {                   /* order here is important */
                              *mult  = tors_mult[j];
                              *force = tors_force[j];
                              *phase = tors_phase[j];
                              *term  = tors_term[j];
                              return(1);
                    }

          if((!strcmp(tors_atom2, tors_name3[j]) &&
              !strcmp(tors_atom3, tors_name2[j])))                      /* two center must match */
                    if((!strcmp(tors_atom1, tors_name4[j])  &&          /* 4 goes to 2 and 1 goes to 3 */
                        !strcmp(tors_atom4, tors_name1[j])) ||
                       ((tors_name1[j][0] == 'X') &&
                        (tors_name4[j][0] == 'X'))) {                   /* order here is important */
                              *mult  = tors_mult[j];
                              *force = tors_force[j];
                              *phase = tors_phase[j];
                              *term  = tors_term[j];
                              return(1);
                    }
     }

     return(0);                                            /* nothing matched */
}




/* need to properly order the impropers before comparisons will work 
   only parameter is pointer to the type by which to order */

void order_impropers(char **assigned_type)
{
     int i, tmp, tmp0, tmp1, tmp3;

     for(i = 0; i < improper_num; i++) {
          tmp0 = improperid[i][0];
          tmp1 = improperid[i][1];
          tmp3 = improperid[i][3];
          if(strcmp(assigned_type[tmp0], assigned_type[tmp1]) > 0) {
               tmp  = tmp1;
               tmp1 = tmp0;
               tmp0 = tmp;
          }
          if(strcmp(assigned_type[tmp0], assigned_type[tmp3]) > 0) {
               tmp  = tmp3;
               tmp3 = tmp0;
               tmp0 = tmp;
          }
          if(strcmp(assigned_type[tmp1], assigned_type[tmp3]) > 0) {
               tmp  = tmp3;
               tmp3 = tmp1;
               tmp1 = tmp;
          }
          improperid[i][0] = tmp0;
          improperid[i][1] = tmp1;
          improperid[i][3] = tmp3;
     }

     return;
}

/* comparison of lexigraphically ordered improper to force field
   parameters are:
        atom types of each atom in improper:               impro_atom1
                                                           impro_atom2
                                                           impro_atom3
                                                           impro_atom4
        number of improper parameter sets in force field:  impro_cntFF
        improper multiplier returned:                      mult
        improper force returned:                           force
        improper angle returned:                           phase
        improper term returned:                            term
*/
int compare_improp(char *impro_atom1, char *impro_atom2, char *impro_atom3,
                   char *impro_atom4, int impro_cntFF, int *mult,
                   float *force, float *phase, float *term)
{
     int j, ibrk;
     char tmp1[4], tmp2[4], tmp3[4], tmp4[4];

     ibrk = -1;
     for(j = 0; j < impro_cntFF; j++) {
          strcpy(tmp1, impro_name1[j]);
          strcpy(tmp2, impro_name2[j]);
          strcpy(tmp3, impro_name3[j]);
          strcpy(tmp4, impro_name4[j]);

          if(impro_xcount[j] > 0)
               if(impro_name3[j][0] == 'X')
                    strcpy(tmp3, impro_atom3);           /* handle the case of 'X' in the key atom */

          if(strcmp(impro_atom3, tmp3)) continue;

          if((!strcmp(tmp1, impro_atom1) || (tmp1[0] == 'X')) &&              /* 1 2 4 */
             (!strcmp(tmp2, impro_atom2) || (tmp2[0] == 'X')) &&
             (!strcmp(tmp4, impro_atom4) || (tmp4[0] == 'X'))) {
                    ibrk = j;
                    break;
          }                                             /* end if((!strcmp(tmp1, impro_atom1) etc. */

          if((!strcmp(tmp1, impro_atom2) || (tmp1[0] == 'X')) &&              /* 2 1 4 */
             (!strcmp(tmp2, impro_atom1) || (tmp2[0] == 'X')) &&
             (!strcmp(tmp4, impro_atom4) || (tmp4[0] == 'X'))) {
                    ibrk = j;
                    break;
          }                                             /* end if((!strcmp(tmp1, impro_atom2) etc. */

          if((!strcmp(tmp1, impro_atom2) || (tmp1[0] == 'X')) &&              /* 2 4 1 */
             (!strcmp(tmp2, impro_atom4) || (tmp2[0] == 'X')) &&
             (!strcmp(tmp4, impro_atom1) || (tmp4[0] == 'X'))) {
                    ibrk = j;
                    break;
          }                                             /* end if((!strcmp(tmp1, impro_atom2) etc. */

          if((!strcmp(tmp1, impro_atom4) || (tmp1[0] == 'X')) &&              /* 4 1 2 */
             (!strcmp(tmp2, impro_atom1) || (tmp2[0] == 'X')) &&
             (!strcmp(tmp4, impro_atom2) || (tmp4[0] == 'X'))) {
                    ibrk = j;
                    break;
          }                                             /* end if(!strcmp(tmp1, impro_atom4) etc. */

          if((!strcmp(tmp1, impro_atom4) || (tmp1[0] == 'X')) &&              /* 4 2 1 */
             (!strcmp(tmp2, impro_atom2) || (tmp2[0] == 'X')) &&
             (!strcmp(tmp4, impro_atom1) || (tmp4[0] == 'X'))) {
                    ibrk = j;
                    break;
          }                                             /* end if(!strcmp(tmp1, impro_atom4) etc. */

          if((!strcmp(tmp1, impro_atom1) || (tmp1[0] == 'X')) &&              /* 1 4 2 */
             (!strcmp(tmp2, impro_atom4) || (tmp2[0] == 'X')) &&
             (!strcmp(tmp4, impro_atom2) || (tmp4[0] == 'X'))) {
                    ibrk = j;
                    break;
          }                                             /* end if(!strcmp(tmp1, impro_atom1) etc. */
     }

     if(ibrk < 0) return(0);                            /* none found */

     *mult  = impro_mult[j];
     *force = impro_force[j];
     *phase = impro_phase[j];
     *term  = impro_term[j];
     return(1);
}
