/* Topology Builder        Bruce D. Ray
   IUPUI Physics Dept. NMR Center
   402 N. Blackford St.
   Indianapolis, IN   46202
   USA
             Main program with functions to print usage and version
             information and to handle amber / gaff processing calls.

   NOTICE: Portions of this are derivative work based on study of:
      1. The routines found GROMACS version 3.3.1 written by David
         van der Spoel, Erik Lindahl, Berk Hess, and others and
         copyright (c) 1991-2000, University of Groningen, The Netherlands.
         Copyright (c) 2001-2004, The GROMACS development team

      2. The routines found in antechamber 1.27 particularly in:
         a. program atomtype, version 1.0, dated October, 2001 by
            Junmei Wang, Department of Pharmaceutical Chemistry,
            School of Pharmacy, University of California, San
            Francisco, CA  94143
         b. programs parmchk and prepgen, version 1.0, dated
            October, 2001 by Junmei Wang, Department of
            Pharmaceutical Chemistry, School of Pharmacy,
            University of California, San Francisco, CA  94143
         c. function file rings.c, by Junmei Wang, Department of
            Pharmaceutical Chemistry, School of Pharmacy, University
            of California, San Francisco, CA  94143
         d. program charmgen, version 2.0, dated June, 2004 by Victor
            E. Bazterra, Center for High Performance Computing, University
            of Utah, Salt Lake City, UT  84114
         e. program bondtype, version 1.0, dated October, 2001 by
            Junmei Wang, Department of Pharmaceutical Chemistry,
            School of Pharmacy, University of California,
            San Francisco, CA  94143
         f. function file mol2.c  by Junmei Wang, Department of
            Pharmaceutical Chemistry, School of Pharmacy, University
            of California, San Francisco, CA  94143

      3. Routines found in AMBCONV by Filip Ryjacek, 2002 as distributed
         at the GROMACS web site.

      4. The routines found in Dock 6.1 file dockmol.cpp version dated
         December 2006, Molecular Design Institute, University of
         California, San Francisco, CA  94143
*/

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "compare_FF.h"
#include "judge_bond.h"
#include "judge_types.h"
#include "mainchain.h"
#include "param_tables.h"
#include "readmol2.h"
#include "ring_detect.h"
#include "read_amber.h"
#include "use_amber.h"
#include "use_gaff.h"
#include "adjustments.h"
#include "similars.h"
#include "Tripos_FF.h"
#include "gromacs_FF.h"

#define MAXCHAR 256
#define MAXNAME 128

void usage(char *name);
void do_an_amber(char *ante_name, int use_amber, char *leap_name, int num_atoms,
                 int num_bonds, char *tp_amber, int angle_max, int tors_max);
void version(char *name);

#define MAXATOM 2000
#define NUMFORCEFIELDS 3

int lstart_at;
int res_start_at;

int main(int argc, char *argv[])
{
     char filename[MAXCHAR], inputname[MAXCHAR], itp_name[MAXCHAR];
     char message[MAXCHAR], ambername[MAXCHAR], leap_name[MAXCHAR];
     char logname[MAXCHAR], outmol[MAXCHAR], corr_name[MAXCHAR];
     char residue[6], dummy[6], amberff[10], ante_name[MAXCHAR];
     char gro_name[MAXCHAR], tp_amber[5];
     int num_atoms, num_bonds, i, j, num_rngs, max_rng, use_mine;
     int angle_max, ang_count, tors_max, tors_count, use_amber;
     int num, num2, count1, count2, nmbr, do_rename;
     int FFmol, FFbond, FFangle, FFtors, FFimpro, FFvdw, FFequiv, FFeqlen;
     char *forcefields[NUMFORCEFIELDS] = { "gaff", "amber", "tripos" };          /* lets me add forcefields later */
     int ff_len[NUMFORCEFIELDS] = { 4, 5, 6 };

     if(argc < 7) {
          if(argc < 2){
               fprintf(stderr, "Need parameters to run.\n");
               usage(argv[0]);
               exit(1);
          }
          if(!strncmp(argv[1], "-h", 2) || !strncmp(argv[1], "-H", 2)) {
               usage(argv[0]);
               exit(0);
          }
          if(!strncmp(argv[1], "-v", 2) || !strncmp(argv[1], "-V", 2)) {
               version(argv[0]);
               exit(0);
          }
          fprintf(stderr, "Need parameters to run.\n");
          usage(argv[0]);
          exit(1);
     }

     filename[0] = NULL;
     ante_name[0] = NULL;
     amberff[0] = NULL;
     residue[0] = NULL;
     use_mine = 0;
     use_amber = -1;
     do_rename = 0;
     lstart_at = 1;
     res_start_at = 0;

     for (i = 1; i < argc; i++) {
          if(!strncmp(argv[1], "-h", 2) || !strncmp(argv[1], "-H", 2)) {
               usage(argv[0]);
               continue;
          }

          if(!strncmp(argv[1], "-v", 2) || !strncmp(argv[1], "-V", 2)) {
               version(argv[0]);
               continue;
          }

          if(!strcmp(argv[i], "-n")) {
               if( strlen(argv[(i + 1)]) > (MAXCHAR-10) ) {
                    sprintf(message, "File name, %s, too long.\n", argv[(i + 1)]);
                    my_fatal(FARGS, message);
                    exit(1);
               }
               strncpy(filename, argv[i+1], (MAXCHAR-10));
               filename[(MAXCHAR-10)] = NULL;
               i++;
               continue;
          }

          if(!strcmp(argv[i], "-dir")) {
               if(strlen(argv[(i + 1)]) > (MAXCHAR-40) ) {
                    sprintf(message, "Forcefield home directory name, %s, too long.\n", argv[(i + 1)]);
                    my_fatal(FARGS, message);
                    exit(1);
               }
               strncpy(ante_name, argv[i+1], (MAXCHAR-40));
               ante_name[(MAXCHAR-40)] = NULL;
               i++;
               continue;
          }

          if(!strcmp(argv[i], "-ff")) {
               strncpy(amberff, argv[(i + 1)], 8);
               amberff[8] = NULL;
               i++;
               continue;
          }

          if(!strcmp(argv[i], "-r")) {
               strncpy(residue, argv[(i + 1)], 5);
               residue[5] = NULL;
               i++;
               continue;
          }
          if(!strcmp(argv[i], "-meas")) {
               use_mine = 1;
               continue;
          }
          if(!strcmp(argv[i], "-rename")) {
               do_rename = 1;
               continue;
          }

          if(!strcmp(argv[i], "-last")) {
               if(strspn(argv[(i + 1)], "0123456789") != strlen(argv[(i + 1)])) {
                    my_fatal(FARGS, "Non digit characters in -last specification\n");
                    exit(1);
               }
               sscanf(argv[(i + 1)], "%d", &lstart_at);
               lstart_at++;
               i++;
               continue;
          }
          if(!strcmp(argv[i], "-resnum")) {
               if(strspn(argv[(i + 1)], "0123456789") != strlen(argv[(i + 1)])) {
                    my_fatal(FARGS, "Non digit characters in -last specification\n");
                    exit(1);
               }
               sscanf(argv[(i + 1)], "%d", &res_start_at);
               i++;
               continue;
          }
     }

     if(!filename[0] || !ante_name[0] || !amberff[0]) {
          fprintf(stderr, "Need correct parameters\n");
          usage(argv[0]);
          exit(1);
     }

     if(!res_start_at && (lstart_at != 1)) {
          my_fatal(FARGS,
                   "Must give last residue number used when giving last atom number used.\n");
          exit(1);
     }

     if(res_start_at && (lstart_at == 1)) {
          my_fatal(FARGS,
                   "Must give last atom number used > 1 when giving last residue number used.\n");
          exit(1);
     }

     strcpy(inputname, filename);
     strcat(inputname, ".mol2");
     strcpy(logname, filename);
     strcat(logname, ".log");
     strcpy(outmol, filename);
     strcat(outmol, "MOL.mol2");
     strcpy(itp_name, filename);
     strcat(itp_name, ".top");
     strcpy(gro_name, filename);
     strcat(gro_name, ".gro");

     for(i = 0; i < NUMFORCEFIELDS; i++)                         /* lets me add forcefields later */
          if(!strncmp(amberff, forcefields[i], ff_len[i])) {
               use_amber = i;
               break;
          }

     if(use_amber < 2) {                             /* gaff and amber use antechamber files */
          strcpy(leap_name, ante_name);
          strcat(ante_name, "/dat/antechamber/");
          strcat(leap_name, "/dat/leap/parm/");
          if(use_amber == 1) strcpy(tp_amber, &amberff[5]);
     }

     if((stdlog = fopen(logname, "w")) == NULL) {
          sprintf(message, "Cannot open log file %s\n", logname);
          my_fatal(FARGS, message);
          exit(1);
     }

     readmol2(inputname, &num_atoms, &num_bonds);

     atom_residue_order(num_atoms, residue);

     new_names = atom_name;
     fprintf(stdlog, "Molecule Name %s\n", mol_name);
     fprintf(stdlog, "Model Type %s\n", model_type);
     fprintf(stdlog, "Charge Method %s\n", charge_type);
     fprintf(stdlog, "Has %d atoms and %d bonds\n", num_atoms, num_bonds);

     angle_max = num_bonds;
     find_angles(0, angle_max, num_atoms, num_bonds, &ang_count);
     angle_max = ang_count;
     find_angles(1, angle_max, num_atoms, num_bonds, &ang_count);

     tors_max = num_bonds;
     find_torsions(0, tors_max, num_atoms, num_bonds, &tors_count);
     tors_max = tors_count;
     find_torsions(1, tors_max, num_atoms, num_bonds, &tors_count);

     measure_bondlen(num_bonds);
     measure_angles(angle_max);
     measure_torsions(tors_max);

     max_rng = num_bonds/2;
     detect_ring(num_atoms, num_bonds, max_rng, &num_rngs);
     loc_improper(num_atoms);

     measure_improp(improper_num);

     judge_bond(num_bonds, num_atoms, num_rngs);
     molec_alloc(num_atoms, num_bonds, ang_count, tors_count,
                 improper_num);                                        /* allocate rest of molecule storage */

     set_mainchain(num_atoms);
     if(do_rename) renameit(num_atoms, num_rngs);

     writemol2(outmol, num_atoms, num_bonds, new_names);

     fprintf(stdlog, "\n Atom Record\n");
     fprintf(stdlog,
             "     no.    name new name   type residue  Sym seg atno           coordinates             charge     mass  bonds\n");

     for(i = 0; i < num_atoms; i++) {
          strncpy(dummy, atom_residue[i], 5);
          dummy[5] = NULL;
          fprintf(stdlog, "%7d %8s %8s %6s  %5s   %2s  %2d %3d %10.4f %10.4f %10.4f %9.4f %9.4f  %2d",
                  (i + lstart_at), atom_name[i], new_names[i], atom_type[i],
                  dummy, atom_symbl[i], atom_segno[i], atom_atno[i], atom_x[i],
                  atom_y[i], atom_z[i], atom_charge[i], atom_mass[i],
                  bond_count[i]);
          for(j = 0; j < main_num; j++)
               if(main_chain[j] == i) {
                    fprintf(stdlog, "    main");
                    break;
               }
          fprintf(stdlog, "\n");
     }

     if(num_rngs) {
/* write out a rings found report */
          fprintf(stdlog, "\n===============================================================\n");
          fprintf(stdlog, "----------------------ring property (I)------------------------\n");
	 
          nmbr = 1;
          for(i = 0; i < num_rngs; i++) {
               if(ring_num[i] == 0) continue;
               fprintf(stdlog, "\n        Ring %d\n\n", nmbr++);
               for(j = 0; j < ring_num[i]; j++)
                    fprintf(stdlog, "%5d %5d %5d %5s\n", (i + 1), (j + 1),
                            (ring_atomno[i][j] + lstart_at),
                            new_names[(ring_atomno[i][j])] );
          }              /* end for(i = 0; i < num_rngs; i++) */

          fprintf(stdlog, "\n===============================================================\n");
          fprintf(stdlog, "---------------------ring property (II)------------------------\n");

          for(i = 0; i < num_rngs; i++) {
               if(ring_num[i] == 0) continue;
               for(j = 0; j < ring_num[i]; j++)
                    fprintf(stdlog,
                            "atom[%5d] (%-4s) belongs to one member of %d-membered ring (No %2d)\n",
                            (ring_atomno[i][j] + lstart_at), new_names[(ring_atomno[i][j])],
                            ring_num[i], (i + 1));
          }
     }

     fprintf(stdlog, "\n===============================================================\n");
     fprintf(stdlog, "---------------------ring property (III)-----------------------\n");

     for(i = 0; i < num_atoms; i++) {
          for(j = 1; j <= 10; j++) {
               if(arom_atomno[i][j] > 0)
                    fprintf(stdlog, "atom[%5d] (%-4s) involved in %d %d-member ring(s)\n",
                            (i + lstart_at), new_names[i], arom_atomno[i][j], j);
          }
          if(arom_num[i] > 0)
               fprintf(stdlog, "atom[%5d] (%-4s) is not in any ring (number[%d]=%d)\n",
                       (i + lstart_at), new_names[i], (i + lstart_at), arom_num[i]);
     }

     fprintf(stdlog, "\n===============================================================\n");
     fprintf(stdlog, "----------------------aromatic property------------------------\n");
     for(i = 0; i < num_atoms; i++) {
          if(ar_set[i][1] >= 1)
               fprintf(stdlog, "atom[%5d] (%-4s) is in %d pure aromatic ring(s) (AR1)\n",
                       (i + lstart_at), new_names[i], ar_set[i][1]);
          if(ar_set[i][2] >= 1)
               fprintf(stdlog, "atom[%5d] (%-4s) is in %d planar ring(s) (AR2)\n",
                       (i + lstart_at), new_names[i], ar_set[i][2]);
          if(ar_set[i][3] >= 1)
               fprintf(stdlog,
                       "atom[%5d] (%-4s) is in %d planar ring(s) that have \"outside\" bonds (AR3)\n",
                       (i + lstart_at), new_names[i], ar_set[i][3]);
          if(ar_set[i][4] >= 1)
               fprintf(stdlog, "atom[%5d] (%-4s) is in %d non-planar ring(s) (AR4)\n",
                       (i + lstart_at), new_names[i], ar_set[i][4]);
          if(ar_set[i][5] >= 1)
               fprintf(stdlog, "atom[%5d] (%-4s) is in %d pure aliphatic ring(s) made of sp3 carbons (AR5)\n",
                       (i + lstart_at), new_names[i], ar_set[i][5]);
     }

     fprintf(stdlog, "\n===============================================================\n");
     fprintf(stdlog, "---------------------electronic property-----------------------\n");
     for(i = 0; i < num_atoms; i++) {
          if(atom_ewd[i])
               fprintf(stdlog, "atom[%5d] (%-4s) is an electron withdrawing atom\n",
                       (i + lstart_at), new_names[i]);
          else
               fprintf(stdlog, "atom[%5d] (%-4s) is not an electron withdrawing atom\n",
                       (i + lstart_at), new_names[i]);
     }

     fprintf(stdlog, "\n===============================================================\n");
     fprintf(stdlog, "--------------------connectivity property----------------------\n");
     for(i = 0; i < num_atoms; i++) {
          for(j = 0; j <= 5; j++) {
               if(connect[i][j] != -1)
                    fprintf(stdlog,  "atom[%5d] (%-4s) %5d %5s\n", (i + lstart_at),
                            new_names[i], (connect[i][j] + lstart_at),
                            new_names[(connect[i][j])]);
          }
     }

     fprintf(stdlog, "------------------------------end------------------------------\n\n");

     fprintf(stdlog, "\nFound %d angles, %d torsions", angle_max, tors_max);
     if(improper_num) fprintf(stdlog, ", and %d impropers\n", improper_num);
     else fprintf(stdlog, "\n");
             
     switch(use_amber) {
                        /* amber force fields */
          case 0:
          case 1:  do_an_amber(ante_name, use_amber, leap_name, num_atoms, num_bonds,
                               tp_amber, angle_max, tors_max);
                   break;
                        /* end amber force fields */

                        /* Tripos FF.  Already have correct atom and bond types */
          case 2:  for(i = 0; i < num_atoms; i++)
                        strcpy(a_ff_type[i], atom_type[i]);
                   FFbond = 10;
                   FFangle = 10;
                   FFtors = 10;
                   FFimpro = 10;
                   FFvdw = 10;

                   read_tripos(ante_name, 0, &FFbond, &FFangle, &FFtors, &FFimpro,
                               &FFvdw);

                   fprintf(stdlog, "\n\tFound\n\t\t%d bond records\n\t\t%d angle records\n",
                           FFbond, FFangle);
                   fprintf(stdlog, "\t\t%d torsion records\n\t\t%d impropers records\n\t\t%d van der Waals records\n",
                           FFtors, FFimpro, FFvdw);

                   read_tripos(ante_name, 1, &FFbond, &FFangle, &FFtors, &FFimpro,
                               &FFvdw);

                   set_amber_vdwFF(num_atoms, FFvdw);                /* works the same for Tripos FF */
                   set_tripos_bondFF(num_bonds, FFbond);
                   set_tripos_angleFF(angle_max, FFangle);
                   set_tripos_torsFF(tors_max, FFtors);
                   set_tripos_improFF(improper_num, FFimpro);
                   free_triposFF(FFbond, FFangle, FFtors, FFimpro, FFvdw);
                   break;
                        /* end Tripos FF */
     }

/* fix measured torsions for multiplicities as needed and able */
     fix_tor(tors_max, meas_tors, a_tors_term, a_tors_force, a_tors_phase);

/* write log of results */
     fprintf(stdlog, "\n\n\tAtom Force Field Results\n");
     fprintf(stdlog, "     no.    name   type   pol.   radius     pot.\n");
     for(i = 0; i < num_atoms; i++) {
          fprintf(stdlog, "%7d %8s", (i + lstart_at),
                  new_names[i]);
          if(a_ff_type[i] == NULL) fprintf(stdlog, "   ***");
          else fprintf(stdlog, " %6s", a_ff_type[i]);
          if(a_pol[i] > 9999.0) fprintf(stdlog, " *******");
          else fprintf(stdlog, " %7.3f", a_pol[i]);
          if(a_vdw_atom_radius[i] > 9999.0) fprintf(stdlog, " ******** ********\n");
          else fprintf(stdlog, " %8.4f %8.4f\n", a_vdw_atom_radius[i],
                       a_vdw_atom_pot[i]);
     }

     fprintf(stdlog, "\n\tBond Force Field Results\n");
     fprintf(stdlog, "  no.     bond    type  interp.  force  length   measured\n");
     for(i = 0; i < num_bonds; i++) {
          fprintf(stdlog, "%4d %5s-%5s   %2s   %3d",
                  (i + 1), new_names[(bond_i[i])], new_names[(bond_j[i])],
                  bond_type[i], how_bonded[i]);
          if(a_bond_force[i] > 9999.0) fprintf(stdlog, "     ****** *******");
          else fprintf(stdlog, "     %6.1f %7.3f",
                       a_bond_force[i], a_bond_length[i]);
          fprintf(stdlog, "    %7.3f\n", meas_len[i]);
     }

     fprintf(stdlog, "\n\tAngles Force Field Results\n");
     fprintf(stdlog, "Angle        Atoms        force    angle     method measured\n");
     for(i = 0; i < angle_max; i++) {
          fprintf(stdlog,"%4d %5s-%5s-%5s", (i + 1), new_names[(angle_tab[i][0])],
                          new_names[(angle_tab[i][1])],
                          new_names[(angle_tab[i][2])]);
          if(a_angle_force[i] > 9999.0) fprintf(stdlog, "   ******   ******   ******");
          else fprintf(stdlog, "   %6.1f   %6.2f   %6d",
                       a_angle_force[i], a_angle_angle[i], a_meth[i]);
          fprintf(stdlog, "    %7.3f\n", meas_ang[i]);
     }

     fprintf(stdlog, "\n\tTorsions Force Field Results\n");
     fprintf(stdlog,
          "Tors           Atoms           mul force   phase   term   measured\n");
     for(i = 0; i < tors_max; i++) {
          fprintf(stdlog,"%4d %5s-%5s-%5s-%5s", (i + 1), new_names[(torsion_tab[i][0])],
                          new_names[(torsion_tab[i][1])],
                          new_names[(torsion_tab[i][2])],
                          new_names[(torsion_tab[i][3])]);
          if(a_tors_mult[i] > 999) fprintf(stdlog,"  ***  ****** ****** ******");
          else fprintf(stdlog, "  %3d %6.2f  %6.1f %6.1f",
                       a_tors_mult[i], a_tors_force[i],
                       a_tors_phase[i], a_tors_term[i]);
         fprintf(stdlog, "   %7.3f\n", meas_tors[i]);
     }

     if(improper_num) {
          fprintf(stdlog, "\n\tImproper Force Field Results\n");
          fprintf(stdlog, "  no.       Improper Locus       force  phase   measured\n");
          for(i = 0; i < improper_num; i++) {
               fprintf(stdlog, "%4d %5s-%5s-%5s-%5s", (i + 1),
                       new_names[(improperid[i][0])], new_names[(improperid[i][1])],
                       new_names[(improperid[i][2])], new_names[(improperid[i][3])]);
               if(a_impr_mult[i] > 999) {
                    fprintf(stdlog, "   ****** ******");
               }
               else
                    fprintf(stdlog, "   %6.1f %6.1f",
                            a_impr_force[i], a_impr_phase[i]);
               fprintf(stdlog, "   %7.3f", meas_impro[i]);
               if(a_impr_mult[i] == 101)
                    fprintf(stdlog, "   default values\n");
               else  
                    fprintf(stdlog, "\n");
          }
     }

/* After final arrays are fixed, set up and write the .gro file */

     write_gro(gro_name, num_atoms, new_names, residue);

/* Then set up and write the .itp file */

     if(use_mine)
          write_top(itp_name, residue, num_atoms, num_bonds, tors_max, angle_max,
                    tors_max, improper_num, a_ff_type, new_names, a_bond_force,
                    meas_len, a_angle_force, meas_ang, a_tors_force, meas_tors,
                    a_tors_term, a_tors_mult, a_impr_force, meas_impro,
                    a_impr_term, a_impr_mult, 1);
     else
          write_top(itp_name, residue, num_atoms, num_bonds, tors_max, angle_max,
                    tors_max, improper_num, a_ff_type, new_names, a_bond_force,
                    a_bond_length, a_angle_force, a_angle_angle, a_tors_force,
                    a_tors_phase, a_tors_term, a_tors_mult, a_impr_force,
                    a_impr_phase, a_impr_term, a_impr_mult, 1);


     write_ff(filename, a_ff_type, num_atoms, a_vdw_atom_radius,
              a_vdw_atom_pot, 1);

     return(0);
}

/* print the usage message
   parameter is the name of the program
*/
void usage(char *name)
{
     version(name);
     printf("Read a Tripos .mol2 file and generate something approximating a gromacs .gro and .itp from it.\n");
     printf("\tNeeds the following parameters and flags:\n");
     printf("\t\t-n                Common name for the .mol2, .gro, .itp, .top, and .log files\n");
     printf("\t\t                  (the path to and name of the input file stripped of the .mol2).\n");
     printf("\t\t                  A revised .mol2 file will be output to common nameMOL.mol2 as well.\n");
     printf("\t\t-dir              Path to a directory that contains appropriate forcefield data.\n");
     printf("\t\t                  For amber type forcefields this would be a pointer to a directory that\n");
     printf("\t\t                  contains directories /dat/antechamber and /dat/leap/parm\n");
     printf("\t\t                  (these in turn must contain the appropriate antechamber and leap data files)\n");
     printf("\t\t                  For tripos forcefield this would be the value of Tripos variable $TA_ASCTABLES\n");
     printf("\t\t-ff               Force field to use, for the time being, either tripos, gaff, or amber##\n");
     printf("\t\t                  where ## designates the year of the amber file (the number between parm\n");
     printf("\t\t                  and the .dat in /dat/leap/parm)\n");
     printf("\t\t-r                Optional residue name to be used for all atoms in the output.\n");
     printf("\t\t-last             Optional last number of a topology to which this topology is to be\n");
     printf("\t\t                  appended.  The first number for an atom in this topology will be one more\n");
     printf("\t\t                  than the number given.  A -resnum is required along with this option.\n");
     printf("\t\t-resnum           Optional last residue number of a topology to which this topology is to be\n");
     printf("\t\t                  appended.  The first number for a residue in this topology will be one more\n");
     printf("\t\t                  than the number given.  A -last is required along with this option.\n");
     printf("\t\t-meas             Optional flag to use measured bond lengths, angles and torsions if set.\n");
     printf("\t\t                  Otherwise, use the values found in the selected forcefield.\n");
     printf("\t\t-rename           Optional flag to rename atoms for the output.\n\n");
     printf("\t\t-h                Prints program version information and this usage message.\n");
     printf("\t\t-v                Prints program version information\n");
     return;
}

/* print the version message
   parameter is the name of the program
*/
void version(char *name)
{
     printf("\n\t\t\t%s\n", name);
     printf("\t\t\tVersion 1.0\n");
     printf("\t\t\tWritten by Bruce D. Ray, Ph.D.\n");
     printf("\t\t\tIUPUI Physics Department\n\t\t\tNMR Center\n");
     printf("\t\t\t402 N. Blackford St.\n");
     printf("\t\t\tIndianapolis, Indiana  46202\n\t\t\tUSA\n\n");
     return;
}

/* Handle amber type force field conversions.
   parameters are:
        ante_name                  directory of antechamber atom type data files
        use_amber                  flag for amber versus gaff force field
        leap_name                  directory of leap amber and gaff parm files
        num_atoms                  number of atoms in molecule
        num_bonds                  number of bonds in molecule
        tp_amber                   year number of amber force field to use
        angle_max                  number of angles in molecule
        tors_max                   number of torsions in molecule 
*/
void do_an_amber(char *ante_name, int use_amber, char *leap_name, int num_atoms,
                 int num_bonds, char *tp_amber, int angle_max, int tors_max)
{
     int num, num2, count1, count2, i, j;
     int FFmol, FFbond, FFangle, FFtors, FFimpro, FFvdw, FFequiv, FFeqlen;
     char ambername[MAXCHAR], corr_name[MAXCHAR];

     strcpy(ambername, ante_name);
     if(use_amber == 1)                                            /* lets me add forcefields later */
          strcat(ambername, "ATOMTYPE_AMBER.DEF");
     if(!use_amber) strcat(ambername, "ATOMTYPE_GFF.DEF");

     read_at_types(&num, &num2, MAXCHAR, MAXCHAR, 0, ambername);     /* first count lines */

     fprintf(stdlog, "\nIn %s there are %d WILDATOM lines and %d ATD lines\n",
                     ambername, num2, num);
     count1 = num;
     count2 = num2;
     read_at_types(&num, &num2, count1, count2, 1, ambername);     /* now read them */

     init_atom_types(num_atoms, num_bonds);
     bond_info(num_atoms, num_bonds);

     if(!use_amber) {
          judge_atom_types(count1, count2, num_atoms, num_bonds, a_ff_type);
          adjustments(num_atoms, num_bonds, a_ff_type);
          adjust_type_cp(num_atoms, num_bonds, a_ff_type);
          check_type_errors(num_atoms, num_bonds, a_ff_type);
     }
     else
          judge_atom_types(count1, count2, num_atoms, num_bonds, a_ff_type);

     free_atom_types(num_atoms);
     release_types(count1, count2);

     if(!use_amber) {
          similars_alloc(num_atoms);                    /* set up the similar names arrays */

          strcpy(corr_name, ante_name);
          strcat(corr_name, "ATCOR.DAT");
          read_at_corr(&num, MAXCHAR, 0, corr_name);     /* count them */

          fprintf(stdlog, "Number of correlation entries in file = %d\n", num);

          num2 = num;
          read_at_corr(&num, num2, 1, corr_name);        /* now read them */

          set_similars(num_atoms, num2);                 /* use them */

          release_corr(num2);                            /* dump them */
     }

     strcpy(ambername, leap_name);
     if(use_amber == 1) {
          strcat(ambername, "parm");
          strcat(ambername, tp_amber);
          strcat(ambername, ".dat");
     }
     if(!use_amber) strcat(ambername, "gaff.dat");

     FFmol = 10;
     FFbond = 10;
     FFangle = 10;
     FFtors = 10;
     FFimpro = 10;
     FFvdw = 10;
     FFequiv = 10;
     FFeqlen = 10;

     read_amber(ambername, 0, &FFmol, &FFbond, &FFangle, &FFtors, &FFimpro,
                &FFvdw, &FFequiv, &FFeqlen);

     fprintf(stdlog, "\n\tFound\n\t\t%d atom records\n\t\t%d bond records\n\t\t%d angle records\n",
            FFmol, FFbond, FFangle);
     fprintf(stdlog, "\t\t%d torsion records\n\t\t%d impropers records\n\t\t%d van der Waals records\n",
            FFtors, FFimpro, FFvdw);
     fprintf(stdlog, "\t\t%d van der Waals equivalence records with %d maximum record length\n",
            FFequiv, FFeqlen);

     FFvdw = FFvdw + (FFequiv * (FFeqlen - 1));
     read_amber(ambername, 1, &FFmol, &FFbond, &FFangle, &FFtors, &FFimpro,
                &FFvdw, &FFequiv, &FFeqlen);

     free_equivs(FFequiv, FFeqlen);

     order_impropers(a_ff_type);

     if(!use_amber) {
          set_gaff_atomFF(num_atoms, FFmol);
          set_gaff_vdwFF(num_atoms, FFvdw);
          set_gaff_bondFF(num_bonds, FFbond);
          set_gaff_angleFF(angle_max, FFangle);
          set_gaff_torsFF(tors_max, FFtors);
          set_gaff_improFF(improper_num, FFimpro);

          for(i = 0; i < num_atoms; i++)                         /* make gaff atom types upper case */
               for(j = 0; j < strlen(a_ff_type[i]); j++)         /* for topology output */
                    a_ff_type[i][j] = toupper(a_ff_type[i][j]);
     }
     else {
          set_amber_atomFF(num_atoms, FFmol);
          set_amber_vdwFF(num_atoms, FFvdw);
          set_amber_bondFF(num_bonds, FFbond);
          set_amber_angleFF(angle_max, FFangle);
          set_amber_torsFF(tors_max, FFtors);
          set_amber_improFF(improper_num, FFimpro);
     }

     return;
}
