/* Topology Builder        Bruce D. Ray
   IUPUI Physics Dept. NMR Center
   402 N. Blackford St.
   Indianapolis, IN   46202
   USA
        Functions to write a simple gromacs format .gro coordinates file
        and to write a gromacs format .top topolgy file

   NOTICE: Portions of this are derivative work based on study of the
   routines found GROMACS version 3.3.1 written by David van der Spoel,
   Erik Lindahl, Berk Hess, and others and copyright (c) 1991-2000,
   University of Groningen, The Netherlands.  Copyright (c) 2001-2004,
   The GROMACS development team
*/

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "gromacs_FF.h"
#include "block_memory.h"
#include "mol2.h"

#define MAXCHAR 256

int lstart_at;
int res_start_at;

/* this is a very simplistic write of a simple gromacs .gro format coordinates file
   that computes the box.
   parameters are:
        filename                 name of .gro file
        atomnum                  number of atoms
        names                    list of atom names
        residue                  the optional residue name for all the atoms
*/
void write_gro (char *filename, int atomnum, char **names, char *residue)
{
     FILE *fpout;
     int i, j, k, count;
     float xmin, ymin, zmin, xmax, ymax, zmax;
     float box_x, box_y, box_z;
     char message[MAXCHAR], dummy[6], *c, *d;

     if((fpout = fopen(filename, "w")) == NULL) {
          sprintf(message, "Cannot open file %s\n", filename);
          my_fatal(FARGS, message);
          exit(1);
     }

     j = 1;
     dummy[0] = NULL;
     fprintf(fpout, "%s\n", mol_name);
     fprintf(fpout, "%5d\n", true_atoms);

     xmin = atom_x[0] *0.1;
     xmax = atom_x[0] *0.1;
     ymin = atom_y[0] *0.1;
     ymax = atom_y[0] *0.1;
     zmin = atom_z[0] *0.1;
     zmax = atom_z[0] *0.1;

     if(residue[0] != NULL) {
          strncpy(dummy, residue, 5);
          dummy[5] = NULL;
     }

     count = 0;
     for(i = 0; i < atomnum; i++) {
          if(atom_order[i] < 0) continue;
          if(residue[0] == NULL) {
               j = atom_segno[(atom_order[i])];
               strncpy(dummy, atom_residue[(atom_order[i])], 5);
               dummy[5] = NULL;
               if((c = strpbrk(dummy, "0123456789")) != NULL) {
                    k = strlen(dummy);
                    if((c != &dummy[0]) &&
                       ((d = strpbrk(&dummy[k-1], "0123456789")) != NULL))
                              *c = NULL;
               }
          }

          fprintf(fpout, "%5d%-5.5s%5.5s%5d%8.3f%8.3f%8.3f\n",
                  (j + res_start_at), dummy, names[(atom_order[i])],
                  ((count + lstart_at)%100000), (atom_x[(atom_order[i])] * 0.1),
                  (atom_y[(atom_order[i])] * 0.1), (atom_z[(atom_order[i])] * 0.1));
          count++;

          if((atom_x[(atom_order[i])] * 0.1) > xmax)
               xmax = atom_x[(atom_order[i])] * 0.1;
          if((atom_y[(atom_order[i])] * 0.1) > ymax)
               ymax = atom_y[(atom_order[i])] * 0.1;
          if((atom_z[(atom_order[i])] * 0.1) > zmax)
               zmax = atom_z[(atom_order[i])] * 0.1;
          if((atom_x[(atom_order[i])] * 0.1) < xmin)
               xmin = atom_x[(atom_order[i])] * 0.1;
          if((atom_y[(atom_order[i])] * 0.1) < ymin)
               ymin = atom_y[(atom_order[i])] * 0.1;
          if((atom_z[(atom_order[i])] * 0.1) < zmin)
               zmin = atom_z[(atom_order[i])] * 0.1;
     }

     box_x = xmax - xmin;
     box_y = ymax - ymin;
     box_z = zmax - zmin;

     fprintf(fpout, "%10.5f %10.5f %10.5f\n", box_x, box_y, box_z);

     fclose(fpout);

     return;
}

/* write the itp file that includes all else
   parameters are:
        filename                     the itp file name for all output
        residue                      the residue name for all the atoms
        numatoms                     number of atoms
        numbonds                     number of bonds
        numpairs                     number of pairs
        numangles                    number of angles
        numdihed                     number of dihedrals
        numimpro                     number of impropers
        assigned_types               list of atom types
        names                        list of atom names
        bforce                       list of bond forces
        blength                      list of bond lengths
        aforce                       list of angle forces
        theta                        list of angle values
        dforce                       list of dihedral forces
        dphi                         list of dihedral phases
        dterm                        list of multiplicity terms to write
        dmul                         list of dihedral multiplicities
        iforce                       list of improper forces
        iphi                         list of improper phases
        iterm                        list of multiplicity terms to write
        imul                         list of improper multiplicities
        amber                        flag for data in kcal/mol and angstroms
*/
void write_top(char *filename, char *residue, int numatoms, int numbonds,
               int numpairs, int numangles, int numdihed, int numimpro,
               char **assigned_types, char **names, float *bforce,
               float *blength, float *aforce, float *theta, float *dforce,
               float *dphi, float *dterm, int *dmul, float *iforce,
               float *iphi, float *iterm, int *imul, int amber)
{
     int i, j, k, impmul, dihmul, tot_rstr, not_pair, itry;
     int rrstr1, rrstr2, count;
     char message[MAXCHAR], *c, *d;
     char dummy[6];
     double sigma_chg, bfc, blen, afc, dfc, ifc, amult;
     double conv = 4.18400000;
     FILE *fptop;

     tot_rstr = rstr_dist + rstr_range + rstr_angle + rstr_torsn;

     if((fptop = fopen(filename, "w")) == NULL) {
          sprintf(message, "Cannot open file %s\n", filename);
          my_fatal(FARGS, message);
          exit(1);
     }

     fprintf(fptop, ";\n; Topology from .mol2 file\n");
     fprintf(fptop, "; topolbuild\n;\n");
     if(!res_start_at) {
          fprintf(fptop, "; The force field files to be included\n");
          strcpy(message, "\"ff");
          strcat(message, filename);
          c = strrchr(message, '.');
          *c = NULL;
          strcat(message, ".itp\"");
          fprintf(fptop, "#include %s\n\n", message);
     }

     dummy[0] = NULL;

     if(residue[0] != NULL) {
          strncpy(dummy, residue, 5);
          dummy[5] = NULL;
     }

     fprintf(fptop, " [ moleculetype ]\n");
     fprintf(fptop, "; name  nrexcl\n%8s   3\n\n", mol_name);

/* atoms */
     fprintf(fptop, " [ atoms ]\n");
     fprintf(fptop, ";  nr    type   resnr   residu   atom   cgnr    charge      mass\n");

     sigma_chg = 0.0;
     j = 1;

     count = 0;
     for(i = 0; i < numatoms; i++) {
          if(atom_order[i] < 0) continue;
          if(residue[0] == NULL) {
               j = atom_segno[(atom_order[i])];
               strncpy(dummy, atom_residue[(atom_order[i])], 5);
               dummy[5] = NULL;
               if((c = strpbrk(dummy, "0123456789")) != NULL) {
                    k = strlen(dummy);
                    if((c != &dummy[0]) &&
                       ((d = strpbrk(&dummy[k-1], "0123456789")) != NULL))
                              *c = NULL;
               }
          }

          sigma_chg += (double)atom_charge[(atom_order[i])];
          fprintf(fptop, "%5d %9s%6d  %8s%8s%5d%10.6f%10.5f       ; %10.6f\n",
                  (count + 1), assigned_types[(atom_order[i])], j, dummy,
                  names[(atom_order[i])], ((i / 4) + 1),
                  atom_charge[(atom_order[i])], atom_mass[(atom_order[i])],
                  sigma_chg);
          count++;
     }

     fprintf(fptop, "; total molecule charge = %10.6f\n", sigma_chg);

/* bonds */
     fprintf(fptop, "\n [ bonds ]\n");
     fprintf(fptop, ";   ai  aj   funct      b0          kb\n");

     for(i = 0; i < numbonds; i++) {
          if((back_order[(bond_i[i])] < 0) ||
             (back_order[(bond_j[i])] < 0))
                    continue;
          blen = (double)blength[i];
          if(amber) blen = blen * 0.1;
          bfc = (double)bforce[i];
          if(amber) bfc = bfc * conv * 200.0;

          fprintf(fptop, "  %6d%6d%s", (back_order[(bond_i[i])] + 1),
                  (back_order[(bond_j[i])] + 1), "   1");
          if(bforce[i] > 9999.0)
               fprintf(fptop, "                                ;%6s-%6s\n", 
                       names[(bond_i[i])], names[(bond_j[i])]);
          else fprintf(fptop, "%12.5f%12.0f%s       ;%6s-%6s\n",
                       blen, bfc, ".", names[(bond_i[i])],
                       names[(bond_j[i])]);
     }

/* treat simple distance restraints as bonds with function type 6, harmonic potential */
     if(rstr_dist) {
          fprintf(fptop, "#ifdef RESTRAIN\n");
          j = rstr_dist;
          for(i = 0; i < tot_rstr; i++) {
               if(restr[i][0] != 1) continue;
               if((back_order[(restr[i][1] - 1)] < 0) ||
                  (back_order[(restr[i][2] - 1)] < 0))
                         continue;
               blen = (double)(frest[i][0]) * 0.1;                      /* these are always amber type */
               bfc = (double)(frest[i][1]) * conv * 200.0;
               fprintf(fptop, "  %6d%6d%s%12.5f%12.0f%s       ;%6s-%6s\n",
                       (back_order[(restr[i][1] - 1)] + 1),
                       (back_order[(restr[i][2] - 1)] + 1),
                       "   6", blen, bfc, names[(restr[i][1] - 1)],
                       names[(restr[i][2] - 1)]);
               j--;
               if(!j) break;
          }
          fprintf(fptop, "#endif\n");
     }

/* pairs */
     fprintf(fptop, "\n [ pairs ]\n");

     for(i = 0; i < numpairs; i++) {
          if((back_order[(torsion_tab[i][0])] < 0) ||
             (back_order[(torsion_tab[i][3])] < 0))
                    continue;
          not_pair = 0;                                     /* must not have a 1 or 2 bond connection elsewhere */
          if((bond_count[(torsion_tab[i][0])] > 1) &&
             (bond_count[(torsion_tab[i][3])] > 1)) {
                    for(j = 0; j < (bond_count[(torsion_tab[i][0])]); j++) {
                         itry = connect[(torsion_tab[i][0])][j];
                         if(bond_count[itry] > 1) {
                              if(itry == torsion_tab[i][3]) {                /* direct bond elsewhere? */
                                   not_pair = 1;
                                   break;
                              }                          /* end if(itry == torsion_tab[i][3])*/

                              for(k = 0; k < bond_count[itry]; k++)          /* only 1 intervening atom? */
                                   if(connect[itry][k] == torsion_tab[i][3]) {
                                        not_pair = 1;
                                        break;
                                   }                      /* end if(connect[itry][k] == torsion_tab[i][3]) */
                              if(not_pair) break;
                         }                              /* end if(bond_count[itry] > 1) */
                    }                                   /* end for(j = 0; j < (bond_count[(torsion_tab[i][0])]); j++) */
          }                                             /* end if((bond_count[(torsion_tab[i][0])] > 1) etc. */

          if(i)                                         /* eliminate duplicate pairs */
               for(j = 0; j < i; j++)
                    if(((torsion_tab[i][0] == torsion_tab[j][0]) &&
                        (torsion_tab[i][3] == torsion_tab[j][3])) || 
                       ((torsion_tab[i][0] == torsion_tab[j][3]) &&
                        (torsion_tab[i][3] == torsion_tab[j][0]))) {
                              not_pair = 1;
                              break;
                    }

          if(not_pair) continue;
          fprintf(fptop, "%8d%8d  1       ;%6s-%6s\n",
                  (back_order[(torsion_tab[i][0])] + 1),
                  (back_order[(torsion_tab[i][3])] + 1),
                  names[(torsion_tab[i][0])], names[(torsion_tab[i][3])]);
     }

/* angles */
     fprintf(fptop, "\n[ angles ]\n");
     fprintf(fptop, "; ai  aj  ak  funct      th0         cth\n");

     for(i = 0; i < numangles; i++) {
          if((back_order[(angle_tab[i][0])] < 0) ||
             (back_order[(angle_tab[i][1])] < 0) ||
             (back_order[(angle_tab[i][2])] < 0))
                    continue;
          afc = (double)aforce[i];
          if(amber) afc = afc * conv * 2.0;

          fprintf(fptop, "%6d%6d%6d%s", (back_order[(angle_tab[i][0])] + 1),
                  (back_order[(angle_tab[i][1])] + 1),
                  (back_order[(angle_tab[i][2])] + 1), "   1");
          if(aforce[i] > 9999.0)
               fprintf(fptop, "                             ;%6s-%6s-%6s\n",
                       names[(angle_tab[i][0])], names[(angle_tab[i][1])],
                       names[(angle_tab[i][2])]);
          else fprintf(fptop, "%12.3f%12.4f     ;%6s-%6s-%6s\n", theta[i], afc,
                       names[(angle_tab[i][0])], names[(angle_tab[i][1])],
                       names[(angle_tab[i][2])]);
     }

/* dihedrals */
     fprintf(fptop, "\n[ dihedrals ]\n");
     fprintf(fptop, "; ai  aj   ak  al  funct    phi0          cp      mult\n");

     for(i = 0; i < numdihed; i++) {
          if((back_order[(torsion_tab[i][0])] < 0) ||
             (back_order[(torsion_tab[i][1])] < 0) ||
             (back_order[(torsion_tab[i][2])] < 0) ||
             (back_order[(torsion_tab[i][3])] < 0))
                    continue;
          dihmul = abs(((int)dterm[i]));
          amult = (double)(abs(dmul[i]));
          if(dihmul > 10) dihmul = 1;
          dfc = (double)dforce[i];
          if(amber) dfc = (dfc * conv) / amult;

          fprintf(fptop, "%6d%6d%6d%6d%s", (back_order[(torsion_tab[i][0])] + 1),
                  (back_order[(torsion_tab[i][1])] + 1),
                  (back_order[(torsion_tab[i][2])] + 1),
                  (back_order[(torsion_tab[i][3])] + 1), "   1");
          if(abs(dmul[i]) > 999) 
               fprintf(fptop, "                                    ; dih %6s-%6s-%6s-%6s\n",
                       names[(torsion_tab[i][0])], names[(torsion_tab[i][1])],
                       names[(torsion_tab[i][2])], names[(torsion_tab[i][3])]);
          else fprintf(fptop, "%12.3f%12.3f%10d  ; dih %6s-%6s-%6s-%6s\n",
                       dphi[i], dfc, dihmul, names[(torsion_tab[i][0])],
                       names[(torsion_tab[i][1])], names[(torsion_tab[i][2])],
                       names[(torsion_tab[i][3])]);
     }

/* impropers, just like dihedrals but use function 2 and no multiplicities. */
     fprintf(fptop, "\n[ dihedrals ]\n");
     fprintf(fptop, ";  ai    aj    ak    al funct           c0           c1\n");

     for(i = 0; i < numimpro; i++) {
          if((back_order[(improperid[i][0])] < 0) ||
             (back_order[(improperid[i][1])] < 0) ||
             (back_order[(improperid[i][2])] < 0) ||
             (back_order[(improperid[i][3])] < 0))
                    continue;
          ifc = (double)iforce[i];
          if(amber) ifc = ifc * conv;

          fprintf(fptop, "%6d%6d%6d%6d%s", (back_order[(improperid[i][2])] + 1),
                  (back_order[(improperid[i][0])] + 1),
                  (back_order[(improperid[i][1])] + 1),
                  (back_order[(improperid[i][3])] + 1), "   2");     /* change from amber order */
          if(abs(imul[i]) > 999) 
               fprintf(fptop, "                                    ; imp %6s-%6s-%6s-%6s\n",
                       names[(improperid[i][2])], names[(improperid[i][0])],
                       names[(improperid[i][1])], names[(improperid[i][3])]);
          else fprintf(fptop, "%12.3f%12.3f            ; imp %6s-%6s-%6s-%6s\n",
                       iphi[i], ifc, names[(improperid[i][2])],
                       names[(improperid[i][0])], names[(improperid[i][1])],
                       names[(improperid[i][3])]);
     }

/* write restraints information */
     if((tot_rstr - rstr_dist)) {
          fprintf(fptop, "\n#ifdef RESTRAIN\n");
          if(rstr_range) {
               fprintf(fptop, "[ distance_restraints ]\n");
               fprintf(fptop, ";   ai  aj   type  index  type'  low  up2  up2  fac\n");
               j = rstr_range;
               for(i = 0; i < tot_rstr; i++) {
                    if((back_order[(restr[i][1] - 1)] < 0) ||
                       (back_order[(restr[i][2] - 1)] < 0))
                              continue;
                    if(restr[i][0] != 2) continue;
                    rrstr1 = back_order[(restr[i][1] - 1)] + 1;
                    rrstr2 = back_order[(restr[i][2] - 1)] + 1;
                    fprintf(fptop, "  %6d%6d%s%6d%s%12.5f%12.5f%12.5f%s       ;%6s-%6s\n",
                            rrstr1, rrstr2,  "   1", i, "   1",
                            (frest[i][0] * 0.1), (frest[i][1] * 0.1),
                            ((frest[i][1] * 0.1) + 1.0), "   1.0",
                            names[(restr[i][1] - 1)], names[(restr[i][2] - 1)]);
                    j--;
                    if(!j) break;
               }
          }

          fprintf(fptop, "#endif\n");
     }

     if(!res_start_at) {
/* write position restraints inclusion lines unless writing an addition */
          strcpy(message, "\"posre");
          strcat(message, filename);
          c = strrchr(message, '.');
          *c = NULL;
          strcat(message, ".itp\"");
          fprintf(fptop, "\n; Include Position restraint file\n");
          fprintf(fptop, "; WARNING: Position restraints and distance restraints ought not be done together\n");
          fprintf(fptop, "#ifdef POSRES\n");
          fprintf(fptop, "#include %s\n", message);
          fprintf(fptop, "#endif\n");

/* write water topology inclusion lines unless writing an addition */
          fprintf(fptop, "\n; Include water topology\n");
          fprintf(fptop, "#include \"gaff_spce.itp\"\n");

          fprintf(fptop, "\n#ifdef POSRES_WATER\n");
          fprintf(fptop, "; Position restraint for each water oxygen\n");
          fprintf(fptop, "[ position_restraints ]\n");
          fprintf(fptop, ";  i funct       fcx        fcy        fcz\n");
          fprintf(fptop, "   1    1       1000       1000       1000\n#endif\n");

/* write ions inclusion lines unless writing an addition */
          fprintf(fptop, "\n; Include generic topology for ions\n");
          fprintf(fptop, "#include \"ions_gaff.itp\"\n");

/* write system information lines unless writing an addition */
          fprintf(fptop, "\n [ system ]\n; title from mol2 input\n");
          fprintf(fptop, "%s\n\n", mol_name);

/* write molecules information lines unless writing an addition */
          fprintf(fptop, " [ molecules ]\n; molecule name    nr.\n");
          fprintf(fptop, "%8s           1\n", mol_name);
     }

     fclose(fptop);
     return;
}
